theory Map_To_Mapping imports
  "~~/src/HOL/Library/Mapping"
  "Mapping_Impl"
begin

text {*
  To convert theorems from @{typ "'a \<Rightarrow> 'b option"} to @{typ "('a, 'b) mapping"} using lifting / transfer,
  we first introduce constants for the empty map and map lookup, then apply lifting / transfer,
  and finally eliminate the non-converted constants again.
*}

context begin interpretation lifting_syntax .

definition map_empty :: "'a \<Rightarrow> 'b option"
where [code_unfold]: "map_empty = Map.empty"

declare Mapping.empty.transfer[transfer_rule del]

lemma map_empty_transfer [transfer_rule]:
  "(pcr_mapping A B) map_empty Mapping.empty"
unfolding map_empty_def by(rule Mapping.empty.transfer)

definition map_apply :: "('a \<Rightarrow> 'b option) \<Rightarrow> 'a \<Rightarrow> 'b option"
where [code_unfold]: "map_apply = (\<lambda>m. m)"

lemma eq_map_apply: "m x \<equiv> map_apply m x"
by(simp add: map_apply_def)

lemma map_apply_parametric [transfer_rule]:
  "((A ===> B) ===> A ===> B) map_apply map_apply"
unfolding map_apply_def by(transfer_prover)

lemma map_apply_transfer [transfer_rule]:
  "(pcr_mapping A B ===> A ===> rel_option B) map_apply Mapping.lookup"
  by (auto simp add: pcr_mapping_def cr_mapping_def Mapping.lookup_def map_apply_def)

definition map_update :: "'a \<Rightarrow> 'b option \<Rightarrow> ('a \<Rightarrow> 'b option) \<Rightarrow> ('a \<Rightarrow> 'b option)"
where "map_update x y f = f(x := y)"

lemma map_update_parametric [transfer_rule]:
  assumes [transfer_rule]: "bi_unique A"
  shows "(A ===> rel_option B ===> (A ===> rel_option B) ===> (A ===> rel_option B)) map_update map_update"
unfolding map_update_def[abs_def] by transfer_prover

lift_definition Mapping_update' :: "'a \<Rightarrow> 'b option \<Rightarrow> ('a, 'b) mapping \<Rightarrow> ('a, 'b) mapping"
is map_update parametric map_update_parametric .

lemma Mapping_update'_code [simp, code, code_unfold]:
  "Mapping_update' x None = Mapping.delete x"
  "Mapping_update' x (Some y) = Mapping.update x y"
by(transfer, simp add: map_update_def fun_eq_iff)+

definition map_is_empty :: "('a \<Rightarrow> 'b option) \<Rightarrow> bool"
where "map_is_empty m \<longleftrightarrow> m = Map.empty"

lemma map_is_empty_folds:
  "m = map_empty \<longleftrightarrow> map_is_empty m"
  "map_empty = m \<longleftrightarrow> map_is_empty m"
by(auto simp add: map_is_empty_def map_empty_def)

lemma map_is_empty_transfer [transfer_rule]:
  assumes "bi_total A"
  shows "(pcr_mapping A B ===> op =) map_is_empty Mapping.is_empty"
unfolding map_is_empty_def[abs_def] Mapping.is_empty_def[abs_def] dom_eq_empty_conv[symmetric]
by(rule rel_funI)+(auto simp del: dom_eq_empty_conv dest: rel_setD2 rel_setD1 Mapping.keys.transfer[THEN rel_funD, OF assms])

end

ML {*
val map_apply_simproc = 
  Simplifier.simproc_global_i @{theory} "map_apply" [@{term "f x :: 'a option"}]
    (fn ctxt => fn redex => case redex of
      Const (@{const_name map_apply}, _) $ _ $ _ => NONE
    | f $ x => 
      let
        val thy = Proof_Context.theory_of ctxt;
        val cTr = 
          redex
          |> fastype_of
          |> dest_Type
          |> snd |> hd
          |> ctyp_of thy;
        val cTx = ctyp_of thy (fastype_of x);
        val cts = map (SOME o cterm_of thy) [f, x];
      in
        SOME (Drule.instantiate' [SOME cTr, SOME cTx] cts @{thm eq_map_apply})
      end
    | _ => NONE);

(* Hack to recover Trasnfer.transferred function from attribute *)
fun transferred ctxt thm = Transfer.transferred_attribute [] (ctxt, thm) |> snd |> the

fun transferred_mapping ctxt thm =
  let
    val ss = put_simpset HOL_basic_ss (Context.proof_of ctxt)
    val ctxt1 = ss addsimps 
      ([@{thm map_empty_def[symmetric]}, 
        @{thm map_update_def[symmetric]}] @
       @{thms map_is_empty_folds})
    val ctxt2 = ss addsimprocs [map_apply_simproc]
    val ctxt3 = ss addsimps 
      ([@{thm map_apply_def}, @{thm map_empty_def},
        @{thm map_update_def[abs_def]}, @{thm map_is_empty_def[abs_def]}] @
       @{thms Mapping_update'_code})
  in
    thm
    |> full_simplify ctxt1
    |> full_simplify ctxt2
    |> transferred ctxt
    |> full_simplify ctxt3
  end

val transferred_mapping_attribute = Thm.rule_attribute transferred_mapping
*}

attribute_setup "transferred_mapping" =
  {* Scan.succeed transferred_mapping_attribute *}
  "Transfer.transferred with processing for ('a, 'b) mapping"

hide_const (open) map_apply map_empty map_is_empty map_update
hide_fact (open) map_apply_def map_empty_def eq_map_apply

lift_definition mapping_filter :: "('a, 'b) mapping \<Rightarrow> 'a list \<Rightarrow> 'b list" is "List.map_filter" .

lemmas mapping_filter_code [code] = map_filter_simps[transferred_mapping]

definition test :: "(nat \<Rightarrow> int option) \<Rightarrow> nat list \<Rightarrow> int list"
where
  "test f xs = 
  (if f = Map.empty then [] else List.map_filter (f(2 := None)(1 \<mapsto> -1)) xs)"

lift_definition test' :: "(nat, int) mapping \<Rightarrow> nat list \<Rightarrow> int list" is test .

lemmas [code] = test_def[transferred_mapping]

export_code test' in SML


fun iter :: "('a \<Rightarrow> 'a option) \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> 'a option"
where
  "iter m 0 x = Some x"
| "iter m (Suc n) x = Option.bind (m x) (iter m n)"

lift_definition iter' :: "('a, 'a) mapping \<Rightarrow> nat \<Rightarrow> 'a \<Rightarrow> 'a option" is iter .

lemmas [code] = iter.simps[transferred_mapping]

export_code iter' in SML

end