(*
Copyright 2009-2013 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Multihole_Context
imports 
  Trs 
  Util
  SubList
  Term_Impl
  "~~/src/HOL/Library/Monad_Syntax"
begin

section {* multihole-contexts *}

datatype_new (-:'f,-:'v)mctxt =  MVar 'v | MHole | MFun 'f "('f,'v)mctxt list"

lemma
  fixes P :: "('f, 'v)mctxt \<Rightarrow> bool"
  assumes "\<And>x. P(MVar x)" and "P MHole" and "\<And>f cs. (\<And>c. c \<in> set cs \<Longrightarrow> P c) \<Longrightarrow> P(MFun f cs)"
  shows mctxt_induct[case_names MVar MHole MFun, induct type: "mctxt"]:
    "P c" 
using assms by (induct c) auto

fun num_holes :: "('f,'v)mctxt \<Rightarrow> nat" where
  "num_holes (MVar _) = 0"
| "num_holes MHole = 1"
| "num_holes (MFun _ ctxts) = listsum (map num_holes ctxts)"

fun fill_holes :: "('f,'v)mctxt \<Rightarrow> ('f,'v)term list \<Rightarrow> ('f,'v)term" where
  "fill_holes (MVar x) _ = Var x"
| "fill_holes MHole [t] = t"
| "fill_holes (MFun f cs) ts = (Fun f (map (\<lambda> i. fill_holes (cs ! i) 
    (take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ts))) [0 ..< length cs]))"

definition partition_holes :: "'b list \<Rightarrow> ('a, 'c) mctxt list \<Rightarrow> 'b list list" where
  "partition_holes ts cs \<equiv> [take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ts) . i \<leftarrow> [0 ..< length cs] ]"

lemma partition_holes_take_drop:
  assumes i: "i < length cs"
  shows "partition_holes ts cs ! i = take (num_holes (cs!i)) (drop (listsum (map num_holes (take i cs))) ts)" 
  proof -
  from i
  have i': "i < length [0..<length cs]" by auto
  hence i'': "[0..<length cs] ! i = i" by auto
  show ?thesis unfolding partition_holes_def nth_map[OF i'] i'' ..
qed  

lemma partition_holes_len: 
  shows " length (partition_holes ts cs) = length cs"
  unfolding partition_holes_def by auto

lemma partition_holes_step: 
  shows "partition_holes ts (c # cs) = take (num_holes c) ts # partition_holes (drop (num_holes c) ts) cs"
  proof - 
    
    have idxs: "0#[Suc 0 ..< length (c#cs)] = [0..<length(c#cs)]"
      using Suc_length_conv Zero_not_Suc gr0I upt_rec by auto
    have fst: "take (num_holes c) ts = take (num_holes ((c#cs)!0)) (drop (listsum (map num_holes (take 0 (c#cs)))) ts)" (is "_ = ?fst") by simp
    {
      fix i
      have "num_holes c + listsum (map num_holes (take i cs)) = listsum (map num_holes (take (Suc i) (c # cs)))" proof (induct cs, auto) qed
      hence "take (num_holes (cs!i)) (drop (listsum (map num_holes (take i cs))) (drop (num_holes c) ts)) 
        = take (num_holes ((c#cs)!(Suc i))) (drop (listsum (map num_holes (take (Suc i) (c # cs)))) ts)" unfolding drop_drop take_map nat_add_commute by auto 
    } note slice = this
    then have "take (num_holes c) ts # partition_holes (drop (num_holes c) ts) cs 
      = ?fst # [take (num_holes ((c#cs)!(Suc i))) (drop (listsum (map num_holes (take (Suc i) (c # cs)))) ts) . i \<leftarrow> [0 ..< length cs]]" 
      unfolding partition_holes_def by auto
    also have "... = ?fst # [take (num_holes ((c#cs)!i)) (drop (listsum (map num_holes (take i (c # cs)))) ts) . i \<leftarrow> [Suc 0 ..< Suc (length cs)] ]"
      unfolding map_Suc_upt[symmetric] map_map by (metis (lifting) comp_apply)
    also have "... = [take (num_holes ((c#cs)!i)) (drop (listsum (map num_holes (take i (c # cs)))) ts) . i \<leftarrow> 0#[Suc 0 ..< length (c#cs)]]" 
      using list.size(4) Nat.add_0_right add_Suc_right by auto
    also have "... = partition_holes ts (c # cs)" unfolding partition_holes_def idxs by simp
    ultimately show ?thesis by auto
  qed


lemma partition_holes_nil:
  shows " partition_holes [] cs = [[] . c \<leftarrow> cs]"
  proof (induct cs, insert take.take_Nil)
  case (Cons c cs) then show ?case unfolding partition_holes_step take.take_Nil drop.drop_Nil by auto
  next
  case Nil show ?case unfolding partition_holes_def  by auto
qed

lemma partition_holes_map_ctxt:
  assumes l: "length cs = length ds"
  and nh: "\<And> i. i < length cs \<Longrightarrow> num_holes (cs!i) = num_holes (ds!i)"
  shows "partition_holes ts cs = partition_holes ts ds"
using l nh proof(induct cs arbitrary: ds ts)
case Nil then have ds: "ds = []" using length_0_conv by auto
show ?case unfolding ds unfolding partition_holes_def by auto
next
  case (Cons c cs')
  then obtain d ds' where ds: "ds = d#ds'" using length_Suc_conv by metis
  have nh': "\<And> i. i < length cs' \<Longrightarrow> num_holes (cs'!i) = num_holes (ds'!i)" proof -
    fix i
    assume i: "i < length cs'"
    have "cs'!i = (c#cs')!(Suc i)" and "ds'!i = (d#ds')!(Suc i)"
      unfolding nth_Cons_Suc by auto
    then show "num_holes (cs'!i) = num_holes (ds'!i)" 
      by (metis Cons.prems(2) ds i length_Suc_conv less_Suc_eq_0_disj)
  qed
  have nh_c_d: "num_holes c = num_holes d" using Cons.prems(2)[of 0] ds nth_Cons_0 by auto
  have l': "length cs' = length ds'" by (metis Cons.prems(1) ds length_Suc_conv list.inject)

  show ?case unfolding ds partition_holes_step nh_c_d using Cons(1)[OF l' nh'] by auto
qed

lemma concat_partition_holes_upto:
  assumes i: "i \<le> length cs"
  shows "concat [partition_holes ts cs ! j . j \<leftarrow> [0..<i]] = take (listsum [num_holes (cs!j) . j \<leftarrow> [0..<i]]) ts"
  using i proof(induct i arbitrary: ts)
  case (Suc i)
  hence i': "i < length cs" by (metis less_eq_Suc_le)
  hence i'':  "i \<le> length cs" by auto
  show ?case
    unfolding upt_Suc_append[OF le0] map_append concat_append Suc(1)[OF i''] concat.simps append_Nil2
    unfolding listsum_append take_add 
    unfolding list.map(2)
    unfolding partition_holes_take_drop[OF i'] 
    unfolding take_upt_idx[OF i'] map_map o_def by auto
  qed (auto)

lemma concat_partition_holes [simp]:
  assumes nh: "listsum (map num_holes cs) = length ts"
  shows "concat (partition_holes ts cs) = ts"
  proof -
    have nh': "(\<Sum>j\<leftarrow>[0..<length cs]. num_holes (cs ! j)) = length ts" 
      using map_upt_len_conv[of num_holes cs] nh[symmetric] by auto
      
    have triv: "length cs \<le> length cs" by auto
    have ph: "partition_holes ts cs = [partition_holes ts cs ! j . j \<leftarrow> [0..<length cs]]" 
      using partition_holes_len by (metis map_nth)
    then show ?thesis using concat_partition_holes_upto[OF triv, unfolded nh'] 
      by (metis order_refl take_all)
qed

lemma concat_partition_holes_id:
  assumes l: "length sss = length cs" 
  and li: "\<And> i. i < length cs \<Longrightarrow> num_holes (cs!i) = length (sss!i)"
  shows "partition_holes (concat sss) cs = sss"
  using l li proof (induct "length sss" arbitrary: sss cs)
  case (Suc i) 
  obtain c cs' where cs: "cs = c # cs'" 
    by (metis Suc.hyps(2) Suc.prems(1) length_Suc_conv)
  obtain ss sss' where sss: "sss = ss # sss'"
    by (metis Suc.hyps(2) length_Suc_conv)
  from Suc.prems(2)[of 0] have l_ss: "num_holes c = length ss" 
    by (metis Suc.hyps(2) Suc.prems(1) cs nth_Cons_0 sss zero_less_Suc)
  have i: "i = length sss'" 
    by (metis Suc.hyps(2) length_Suc_conv list.inject sss)
  have l': "length sss' = length cs'" 
    by (metis Suc.prems(1) cs length_Suc_conv list.inject sss)
  have li': "\<And> i. i < length cs' \<Longrightarrow> num_holes (cs'!i) = length (sss'!i)"
    by (metis Suc.prems(1) Suc.prems(2) cs length_Suc_conv not_less_eq nth_Cons_Suc sss)
  have "partition_holes (concat (ss # sss')) cs = ss #
    partition_holes (concat sss') cs'"
    unfolding cs sss concat.simps(2) partition_holes_step
    by (metis append_eq_conv_conj l_ss)
  then show ?case using Suc(1)[OF i l' li'] sss by metis 
qed (metis less_nat_zero_code nth_equalityI partition_holes_len)
  


lemma partition_holes_map_nth:
  assumes i: "i < length cs"
  shows "partition_holes [ f ti . ti \<leftarrow> ts] cs ! i = [f ti. ti \<leftarrow> partition_holes ts cs ! i]"
  using i proof (induct i arbitrary: cs ts)
  case 0 
  from i obtain c cs' where cs: "cs = c # cs'" by (metis "0.prems" length_greater_0_conv list.exhaust)
  show ?case unfolding cs partition_holes_step nth_Cons_0 take_map ..
  next

  case (Suc i)
  from i obtain c cs' where cs: "cs = c # cs'" by (metis Suc.prems Suc_length_conv lessE)
  have i: "i < length cs'" using list.size(4) Suc.prems[unfolded cs] by auto

  have left: "partition_holes (map f ts) cs ! (Suc i) = partition_holes (map f (drop (num_holes c) ts)) cs' ! i" 
    unfolding cs partition_holes_step nth_Cons_Suc drop_map ..
  
  have right: "map f (partition_holes ts cs ! (Suc i)) = map f (partition_holes (drop (num_holes c) ts) cs' ! i)" 
    unfolding cs partition_holes_step nth_Cons_Suc ..

  from left right Suc(1)[OF i] show ?case by auto
qed

lemma partition_holes_nth_length:
  assumes nh: "listsum (map num_holes cs) = length ts"
  and i: "i < length cs"
  shows "num_holes (cs!i) = length (partition_holes ts cs!i)"
proof -
  have "cs!i = hd (drop i cs)" using hd_drop_conv_nth[OF i] by auto
  moreover have "drop i cs \<noteq> []" using drop_eq_Nil i using not_le by auto
  ultimately have "cs!i \<in> set (drop i cs)" using hd_in_set by auto
  hence "num_holes (cs!i) \<in> set (map num_holes (drop i cs))" by auto
  hence *: "num_holes (cs!i) \<le> listsum (map num_holes (drop i cs))" using member_le_listsum_nat by auto

  have "length ts = listsum (map num_holes (take i cs) @ map num_holes (drop i cs))"
    unfolding nh[symmetric] num_holes.simps
    map_append[symmetric] append_take_drop_id by simp
  moreover have "... = listsum (map num_holes (take i cs)) + listsum (map num_holes (drop i cs))" using listsum_append by auto
  ultimately have "length (drop (listsum (map num_holes (take i cs))) ts) = listsum (map num_holes (drop i cs))"
    by (metis add_diff_cancel_left' length_drop) 
  from *[unfolded this[symmetric]] partition_holes_def[of ts cs] length_take 
  have **: "num_holes (cs!i) = length (take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ts))" (is "_ = length ?tsi'")
    by (metis min_absorb2)

  from i length_upt have "i < length [0..<length cs]" by auto  
  moreover hence "[0..<length cs] ! i = i" by auto
  ultimately have "partition_holes ts cs!i = ?tsi'" unfolding partition_holes_def 
    using nth_map i by auto 
  then show ?thesis  using ** by auto
qed

lemma partition_holes_fill_holes_conv:
  shows "fill_holes (MFun f cs) ts = Fun f [ fill_holes (cs!i) (partition_holes ts cs!i) . i \<leftarrow> [0..<length cs]]"
  unfolding partition_holes_def by auto

lemma fill_holes_inj: assumes "fill_holes C ss = fill_holes C ts"
  and ss: "num_holes C = length ss"
  and ts: "length ts = length ss"
  shows "ss = ts"
  using assms
proof (induct C arbitrary: ss ts)
  case (MHole ss ts)
  thus ?case by (cases ss, (cases ts, auto)+)
next
  case (MFun ff Cs ss ts)
  let ?s = "\<lambda> C ss. fill_holes C (take (num_holes C) ss)"
  let ?t = "\<lambda> i Cs ss. fill_holes (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) ss))"
  let ?p = "\<lambda> i Cs ss ts. ?t i Cs ss = ?t i Cs ts"
  from MFun(2) have p: "\<And> i. i < length Cs \<Longrightarrow> ?p i Cs ss ts" by auto
  from MFun(3) have nh: "listsum (map num_holes Cs) = length ss" by auto
  from MFun(4) have l: "length ts = length ss" by auto
  from MFun(1) p nh l show ?case
  proof (induct Cs arbitrary: ss ts)
    case (Cons C Cs ss ts)
    note oIH = Cons(2)[of C]
    let ?n = "num_holes C"
    from Cons(3)[of 0] have id: "?s C ss = ?s C ts" by auto      
    from Cons(4) Cons(5) have n: "?n = length (take ?n ss)" "length (take ?n ts) = length (take ?n ss)" by auto
    from oIH[OF _ id n] have id: "take ?n ss = take ?n ts" by auto
    have idss: "\<And> f. f ss = f (take ?n ss @ drop ?n ss)" by simp
    have idts: "\<And> f. f ts = f (take ?n ss @ drop ?n ts)" using append_take_drop_id[of ?n ts, unfolded id[symmetric]] by auto
    have "(ss = ts) \<longleftrightarrow> take ?n ss @ drop ?n ss = take ?n ts @ drop ?n ts" by auto
    also have "\<dots> \<longleftrightarrow> drop ?n ss = drop ?n ts" unfolding id by (simp del: append_take_drop_id)
    also have "\<dots>"
    proof (rule Cons(1))
      fix C ss ts
      assume C: "C \<in> set Cs"
      and *: "fill_holes C ss = fill_holes C ts" "num_holes C = length ss" "length ts = length ss"
      from Cons(2)[OF _ *] C show "ss = ts" by auto
    next
      show "listsum (map num_holes Cs) = length (drop (num_holes C) ss)"
        using Cons(4) n(1) by auto
      show "length (drop (num_holes C) ts) = length (drop (num_holes C) ss)"
        using Cons(5)[unfolded idts[of length] idss[of length]] by auto
    next
      fix i
      assume "i < length Cs"
      hence i: "Suc i < length (C # Cs)" by simp
      show "?p i Cs (drop ?n ss) (drop ?n ts)"
        using Cons(3)[OF i] by (auto simp: ac_simps)
    qed
    finally show ?case by simp
  qed auto
qed auto

lemma fill_holes_arbitrary: assumes lCs: "length Cs = length ts"
  and lss: "length ss = length ts"
  and rec: "\<And> i. i < length ts \<Longrightarrow> num_holes (Cs ! i) = length (ss ! i) \<and> f (Cs ! i) (ss ! i) = ts ! i"
  shows "map (\<lambda>i. f (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) (concat ss)))) [0..<length Cs] 
    = ts"
proof -
  have l: "length [0 ..< length Cs] = length ts" unfolding lCs by simp
  show ?thesis 
  proof (unfold map_nth_eq_conv[OF l], intro allI impI)
    fix i
    assume i: "i < length ts"
    with lCs have iC: "i < length Cs" by simp
    hence "f (Cs ! ([0..<length Cs] ! i))
         (take (num_holes (Cs ! ([0..<length Cs] ! i))) (drop (listsum (map num_holes (take ([0..<length Cs] ! i) Cs))) (concat ss)))
       = f (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) (concat ss)))" (is "?l = _")
     by simp
    also have "\<dots> = ts ! i" using i lCs lss rec
    proof (induct i arbitrary: ts Cs ss)
      case 0
      from 0 obtain t ts' where ts: "ts = t # ts'" by (cases ts, auto)
      from 0 obtain s ss' where ss: "ss = s # ss'" by (cases ss, auto)
      from 0 obtain C Cs' where Cs: "Cs = C # Cs'" by (cases Cs, auto)
      show ?case using 0(4)[of 0] unfolding ts ss Cs by auto
    next
      case (Suc i)
      from Suc obtain t ts' where ts: "ts = t # ts'" by (cases ts, auto)
      from Suc obtain s ss' where ss: "ss = s # ss'" by (cases ss, auto)
      from Suc obtain C Cs' where Cs: "Cs = C # Cs'" by (cases Cs, auto)
      from Suc ts ss Cs have i: "i < length ts'" "length Cs' = length ts'" "length ss' = length ts'" by auto
      from Suc(5)[of 0] ts ss Cs have nh: "num_holes C = length s" "f C s = t" by auto
      show ?case using Suc(5) unfolding Cs ss ts using nh
        by (force intro!: Suc(1) i)
    qed
    finally show "?l = ts ! i" .
  qed
qed

lemma fill_holes_MFun: assumes lCs: "length Cs = length ts"
  and lss: "length ss = length ts"
  and rec: "\<And> i. i < length ts \<Longrightarrow> num_holes (Cs ! i) = length (ss ! i) \<and> fill_holes (Cs ! i) (ss ! i) = ts ! i"
  shows "fill_holes (MFun f Cs) (concat ss) = Fun f ts" 
  unfolding fill_holes.simps term.simps
    by (rule conjI[OF refl], rule fill_holes_arbitrary[OF lCs lss rec])

inductive eq_fill :: "('f,'v)term \<Rightarrow> ('f,'v)mctxt \<times> ('f,'v)term list \<Rightarrow> bool" ("(_/ =f _)" [51,51] 50) where
  eqfI[intro]: "t = fill_holes D ss \<Longrightarrow> num_holes D = length ss \<Longrightarrow> t =f (D,ss)"

lemma eqfE: assumes "t =f (D,ss)" shows "t = fill_holes D ss" "num_holes D = length ss"
  using assms[unfolded eq_fill.simps] by auto

lemma eqf_MFunI: assumes "length sss = length Cs"
  "length ts = length Cs"
  "\<And> i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)"
  shows "Fun f ts =f (MFun f Cs,concat sss)"
proof 
  have "num_holes (MFun f Cs) = listsum (map num_holes Cs)" by simp
  also have "map num_holes Cs = map length sss"
    by (rule nth_equalityI, insert assms eqfE[OF assms(3)], auto)
  also have "listsum (\<dots>) = length (concat sss)" unfolding length_concat ..
  finally show "num_holes (MFun f Cs) = length (concat sss)" .
  show "Fun f ts = fill_holes (MFun f Cs) (concat sss)"
    by (rule fill_holes_MFun[symmetric], insert assms(1,2) eqfE[OF assms(3)], auto)
qed

lemma eqf_MFunE: assumes "s =f (MFun f Cs,ss)"  
  obtains ts sss where "s = Fun f ts" "length ts = length Cs" "length sss = length Cs" 
  "\<And> i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)"
  "ss = concat sss" 
proof -
  from eqfE[OF assms] have fh: "s = fill_holes (MFun f Cs) ss" 
    and nh: "listsum (map num_holes Cs) = length ss" by auto 
  from fh obtain ts where s: "s = Fun f ts" by (cases s, auto)
  from fh[unfolded s] 
  have ts: "ts = map (\<lambda>i. fill_holes (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) ss))) [0..<length Cs]" 
    (is "_ = map (?f Cs ss) _")
    by auto
  from nh 
  have "\<exists> sss. length sss = length Cs \<and> (\<forall> i < length Cs . ts ! i =f (Cs ! i, sss ! i)) \<and> ss = concat sss"
    unfolding ts
  proof (induct Cs arbitrary: ss)
    case (Cons C Cs ss)
    let ?n = "num_holes C"
    let ?bef = "take ?n ss"
    let ?aft = "drop ?n ss"
    def bef \<equiv> ?bef
    def aft \<equiv> ?aft
    from Cons(2) have "length ?bef = ?n" "listsum (map num_holes Cs) = length ?aft" by auto
    hence lb: "?n = length bef" "listsum (map num_holes Cs) = length aft" unfolding bef_def aft_def by auto
    have ss: "ss = bef @ aft" unfolding bef_def aft_def by auto
    have l: "length (C # Cs) = Suc (length Cs)" by simp
    note IH = Cons(1)[OF lb(2)]
    from IH obtain sss where len: "length sss = length Cs" and
      i: "\<And> i. i < length Cs \<Longrightarrow> map (?f Cs aft) [0 ..< length Cs] ! i =f (Cs ! i, sss ! i)" and
      aft: "aft = concat sss" by blast
    have "?case \<longleftrightarrow> (\<exists>sss. length sss = Suc (length Cs) \<and>
          fill_holes C bef =f (C, sss ! 0) \<and>
          (\<forall>i<length Cs.
              ?f Cs aft i =f (Cs ! i, sss ! Suc i)) \<and>
          bef @ aft = concat sss)"
      unfolding l map_upt_Suc all_Suc_conv 
      unfolding ss by (simp add: lb)
    also have \<dots>
    proof (rule exI[of _ "bef # sss"], intro conjI allI impI)
      show "length (bef # sss) = Suc (length Cs)" by (simp add: len)
      show "fill_holes C bef =f (C, (bef # sss) ! 0)" using eqfI[OF _ lb(1)] by simp
      show "bef @ aft = concat (bef # sss)" using aft by simp
      fix i
      assume "i < length Cs"
      with i[OF this]
      show "?f Cs aft i =f (Cs ! i, (bef # sss) ! Suc i)" by simp
    qed        
    finally show ?case by simp
  qed auto
  then obtain sss where *: "length sss = length Cs" "\<And> i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)"
    "ss = concat sss" by blast
  have len: "length ts = length Cs" unfolding ts by auto
  assume ass: "\<And>ts sss. s = Fun f ts \<Longrightarrow>
              length ts = length Cs \<Longrightarrow>
              length sss = length Cs \<Longrightarrow> (\<And>i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)) \<Longrightarrow> ss = concat sss \<Longrightarrow> thesis"
  show thesis
    by (rule ass[OF s len *])
qed

lemma fill_holes_ctxt_main: assumes "num_holes C = Suc (length bef + length aft)"
  shows "\<exists> D. \<forall> s. fill_holes C (bef @ s # aft) = D \<langle> s \<rangle>"
  using assms
proof (induct C arbitrary: bef aft)
  case MHole
  show ?case
    by (rule exI[of _ \<box>], insert MHole, auto)
next
  case (MFun f cs)
  note IH = MFun(1)
  note holes = MFun(2)
  let ?p = "\<lambda> bef aft b a D cs s. map (\<lambda>i. fill_holes (cs ! i)
                        (take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) (bef @ s # aft))))
               [0..<length cs] =
              b @ D\<langle>s\<rangle> # a"
  from holes IH
  have "\<exists> b D a. \<forall>s. ?p bef aft b a D cs s"
  proof (induct cs arbitrary: bef)
    case (Cons c ccs)
    have len: "length (c # ccs) = Suc (length ccs)" by simp
    show ?case 
    proof (cases "num_holes c \<le> length bef")
      case True
      hence "bef = take (num_holes c) bef @ drop (num_holes c) bef 
        \<and> length (take (num_holes c) bef) = num_holes c" by auto
      then obtain bc ba where bef: "bef = bc @ ba" and lbc: "length bc = num_holes c" by blast
      from Cons(2) have nh: "num_holes (MFun f ccs) = Suc (length ba + length aft)" unfolding bef
        by (simp add: lbc)
      from Cons(1)[OF nh Cons(3)] obtain b D a where IH: "\<And> s. ?p ba aft b a D ccs s" by auto
      show ?thesis unfolding len map_upt_Suc bef
        by (rule exI[of _ "fill_holes c bc # b"], rule exI[of _ D], rule exI[of _ a], insert IH lbc, auto)
    next
      case False
      hence "\<exists> la. num_holes c = Suc (length bef + la)" by arith
      then obtain la where nhc: "num_holes c = Suc (length bef + la)" ..
      from Cons(2) nhc have "length (take la aft) = la" by auto
      from Cons(3)[of c bef "take la aft", unfolded this, OF _ nhc]
      obtain D where D: "\<forall>s. fill_holes c (bef @ s # take la aft) = D\<langle>s\<rangle>" by auto
      show ?thesis unfolding len map_upt_Suc
        by (rule exI[of _ Nil], rule exI[of _ D], simp add: nhc D) 
    qed
  qed auto
  then obtain b D a where main: "\<And> s. ?p bef aft b a D cs s" by blast
  show ?case by (rule exI[of _ "More f b D a"], insert main, auto)
qed simp

lemma fill_holes_ctxt: assumes nh: "num_holes C = length ss"
  and i: "i < length ss"
  obtains D where "\<And> s. fill_holes C (ss[i := s]) = D \<langle> s \<rangle>"
proof -
  from id_take_nth_drop[OF i] obtain bef aft where ss: "ss = bef @ ss ! i # aft" 
    and bef: "bef = take i ss" by blast
  from bef i have bef: "length bef = i" by auto
  note len = arg_cong[OF ss, of length]
  from len nh
  have "num_holes C = Suc (length bef + length aft)" by simp
  from fill_holes_ctxt_main[OF this] obtain D where id: "\<And> s. fill_holes C (bef @ s # aft) = D \<langle> s \<rangle>" by blast
  {
    fix s
    have "ss[i := s] = bef @ s # aft" unfolding arg_cong[OF ss, of "\<lambda> ss. ss[i := s]"]
      using i bef by auto
    with id[of s] have "fill_holes C (ss[i := s]) = D \<langle>s \<rangle>" by simp
  }
  hence main: "\<exists> D. \<forall> s. fill_holes C (ss[i := s]) = D \<langle>s \<rangle>" by blast
  assume "\<And> D. \<lbrakk>\<And>s. fill_holes C (ss[i := s]) = D\<langle>s\<rangle>\<rbrakk> \<Longrightarrow> thesis"
  with main show thesis by blast
qed

fun map_vars_mctxt :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f,'v)mctxt \<Rightarrow> ('f,'w)mctxt" where
  "map_vars_mctxt vw MHole = MHole"
| "map_vars_mctxt vw (MVar v) = (MVar (vw v))"
| "map_vars_mctxt vw (MFun f Cs) = MFun f (map (map_vars_mctxt vw) Cs)"

lemma map_vars_mctxt_id[simp]: "map_vars_mctxt (\<lambda> x. x) C = C"
  by (induct C, auto intro: nth_equalityI)

lemma num_holes_map_vars_mctxt[simp]: "num_holes (map_vars_mctxt vw C) = num_holes C"
proof (induct C)
  case (MFun f Cs)
  thus ?case by (induct Cs, auto)
qed auto

lemma map_vars_eq_fill: "t =f (C,ss) \<Longrightarrow> map_vars vw t =f (map_vars_mctxt vw C, map (map_vars vw) ss)"
proof (induct C arbitrary: t ss)
  case (MFun f Cs s ss)
  from eqf_MFunE[OF MFun(2)] obtain ts sss where s: "s = Fun f ts" and len: "length ts = length Cs" "length sss = length Cs" 
    and IH: "\<And> i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)" and ss: "ss = concat sss" by metis
  {
    fix i
    assume i: "i < length Cs"
    hence "Cs ! i \<in> set Cs" by auto
    from MFun(1)[OF this IH[OF i]] have "map_vars vw (ts ! i) =f (map_vars_mctxt vw (Cs ! i), map (map_vars vw) (sss ! i))" .
  } note IH = this
  show ?case unfolding map_vars_mctxt.simps ss map_concat s map_vars.simps
    by (rule eqf_MFunI, insert IH len, auto)
next
  case (MHole t ss)
  from eqfE[OF this]
  show ?case by (cases ss, auto)
next
  case (MVar v t ss)
  from eqfE[OF this]
  show ?case by (cases ss, auto)
qed

lemma map_vars_fill_holes: assumes nh: "num_holes C = length ss"
  shows "map_vars vw (fill_holes C ss) = fill_holes (map_vars_mctxt vw C) (map (map_vars vw) ss)"
proof -
  from eqfE[OF map_vars_eq_fill[OF eqfI[OF refl nh]]]
  show ?thesis by simp
qed


fun split_vars :: "('f,'v)term \<Rightarrow> (('f,'v)mctxt \<times> 'v list)" where
  "split_vars (Var x) = (MHole, [x])"
| "split_vars (Fun f ts) = (MFun f (map fst (map split_vars ts)), concat (map snd (map split_vars ts)))"

lemma split_vars_num_holes: "num_holes (fst (split_vars t)) = length (snd (split_vars t))"
proof (induct t)
  case (Fun f ts)
  thus ?case by (induct ts, auto)
qed simp

lemma split_vars_vars_term_list': "snd (split_vars t) = vars_term_list' t"
proof (induct t)
  case (Fun f ts)
  thus ?case by (auto simp: vars_term_list'.simps o_def, induct ts, auto)
qed (auto simp: vars_term_list'.simps)


lemma split_vars_vars_term: "set (snd (split_vars t)) = vars_term t"
  using arg_cong[OF split_vars_vars_term_list'[of t], of set] by auto

lemma split_vars_eqf_subst_map_vars: "t \<cdot> \<sigma> =f (map_vars_mctxt vw (fst (split_vars t)), map \<sigma> (snd (split_vars t)))"
proof (induct t)
  case (Fun f ts)
  have "?case = (Fun f (map (\<lambda>t. t \<cdot> \<sigma>) ts)
    =f (MFun f (map (map_vars_mctxt vw \<circ> (fst \<circ> split_vars)) ts), concat (map (map \<sigma> \<circ> (snd \<circ> split_vars)) ts)))"
    by (simp add: map_concat)
  also have "..." 
  proof (rule eqf_MFunI, simp, simp, unfold length_map)
    fix i
    assume i: "i < length ts"
    hence mem: "ts ! i \<in> set ts" by auto
    show "map (\<lambda>t. t \<cdot> \<sigma>) ts ! i =f (map (map_vars_mctxt vw \<circ> (fst \<circ> split_vars)) ts ! i, map (map \<sigma> \<circ> (snd \<circ> split_vars)) ts ! i)"
      using Fun[OF mem] i by auto
  qed
  finally show ?case by simp
qed auto

lemma split_vars_eqf_subst: "t \<cdot> \<sigma> =f (fst (split_vars t), (map \<sigma> (snd (split_vars t))))"
  using split_vars_eqf_subst_map_vars[of t \<sigma> "\<lambda> x. x"] by simp

lemma split_vars_into_subst_map_vars: assumes split: "split_vars l = (C,xs)"
  and len: "length ts = length xs"
  and id: "\<And> i. i < length xs \<Longrightarrow> \<sigma> (xs ! i) = ts ! i"
  shows "l \<cdot> \<sigma> =f (map_vars_mctxt vw C,ts)"
proof -
  from split_vars_eqf_subst_map_vars[of l \<sigma> vw, unfolded split]
  have "l \<cdot> \<sigma> =f (map_vars_mctxt vw C, map \<sigma> xs)" by simp
  also have "map \<sigma> xs = ts"
    by (rule nth_equalityI, insert len id, auto)
  finally show ?thesis .
qed

lemma split_vars_into_subst: assumes split: "split_vars l = (C,xs)"
  and len: "length ts = length xs"
  and id: "\<And> i. i < length xs \<Longrightarrow> \<sigma> (xs ! i) = ts ! i"
  shows "l \<cdot> \<sigma> =f (C,ts)"
  using split_vars_into_subst_map_vars[OF split len id, of "\<lambda> x. x"] by simp

fun funas_mctxt :: "('f,'v)mctxt \<Rightarrow> 'f sig" where 
  "funas_mctxt (MFun f Cs) = insert (f, length Cs) (\<Union>(funas_mctxt ` (set Cs)))"
| "funas_mctxt _ = {}"

lemma eqf_funas_term:
  "t =f (C,ss) \<Longrightarrow> funas_term t = funas_mctxt C \<union> \<Union>(funas_term ` set ss)"
proof (induct C arbitrary: t ss)
  case (MFun f Cs t ss)
  from eqf_MFunE[OF MFun(2)] obtain ts sss where
    t: "t = Fun f ts" and len: "length ts = length Cs" "length sss = length Cs" 
    and args: "\<And> i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)"
    and ss: "ss = concat sss"  by auto
  let ?lhs = "\<Union> {funas_term (ts ! i) | i. i < length Cs}"
  let ?f1 = "\<lambda> i. funas_mctxt (Cs ! i)"
  let ?f2 = "\<lambda> i. \<Union>(funas_term ` set (sss ! i))"
  let ?f = "\<lambda> i. ?f1 i \<union> ?f2 i"
  {
    fix i
    assume i: "i < length Cs"
    hence mem: "Cs ! i \<in> set Cs" by auto
    note MFun(1)[OF mem args[OF i]]
  } note IH = this
  have "funas_term t = insert (f,length Cs) ?lhs"
    unfolding t using len by (auto simp: set_conv_nth)
  also have "?lhs = \<Union> {?f i | i. i < length Cs}" using IH by (metis (hide_lams)) 
  also have "\<dots> = \<Union> {?f1 i | i. i < length Cs} \<union> \<Union> {?f2 i | i. i < length Cs}" by auto
  also have "insert (f,length Cs) \<dots> = (insert (f,length Cs) (\<Union> {?f1 i | i. i < length Cs})) \<union> \<Union> {?f2 i | i. i < length Cs}" by auto
  also have "insert (f,length Cs) (\<Union> {?f1 i | i. i < length Cs}) = funas_mctxt (MFun f Cs)"
    by (auto simp: set_conv_nth)
  also have "\<Union> {?f2 i | i. i < length Cs} = \<Union>(funas_term ` set ss)" unfolding ss len(2)[symmetric] 
    using set_conv_nth[of sss] by auto
  finally show ?case .
next
  case MVar
  from eqfE[OF this]
  show ?case by auto
next
  case MHole
  from eqfE[OF this] show ?case by (cases ss, auto)
qed 

lemma eqf_all_ctxt_closed_step: assumes ctxt: "all_ctxt_closed F R" and 
  ass: "t =f (D,ss)" "\<And> i. i < length ts \<Longrightarrow> (ss ! i, ts ! i) \<in> R" "length ss = length ts" "funas_term t \<subseteq> F" 
  "UNION (set ts) funas_term \<subseteq> F"
  shows "(t,fill_holes D ts) \<in> R \<and> fill_holes D ts =f (D,ts)" 
proof -
  have R: "(t,fill_holes D ts) \<in> R" using ass
  proof (induct D arbitrary: t ss ts)
    case (MVar x)
    from eqfE[OF MVar(1)] have id: "t = Var x" "ss = []" by auto
    with MVar(3) have ts: "ts = []" by auto
    show ?case unfolding id ts using all_ctxt_closed_sig_reflE[OF ctxt] by auto
  next
    case MHole
    from eqfE[OF MHole(1)] have id: "ss = [t]" by (cases ss, auto)
    with MHole(3) obtain t' where ts: "ts = [t']" by (cases ts, auto)
    show ?case using MHole(2)[of 0] unfolding id ts by auto
  next
    case (MFun f Cs)
    let ?fh = "fill_holes (MFun f Cs)"
    note eqft = MFun(2)
    from MFun(5)[unfolded eqf_funas_term[OF eqft]] MFun(6)
    have F: "(f, length Cs) \<in> F" "UNION (set Cs) funas_mctxt \<subseteq> F" "UNION (set ss) funas_term \<subseteq> F" "UNION (set ts) funas_term \<subseteq> F" by auto
    from eqfE[OF eqft] have t: "t = ?fh ss" and nh: "listsum (map num_holes Cs) = length ss" by auto    
    have length_upt: "\<And> x. length [0 ..< x] = x" by auto
    show ?case unfolding t fill_holes.simps
    proof (rule all_ctxt_closedE[OF ctxt], unfold length_map length_upt, rule F(1), simp)
      let ?list = "\<lambda> Cs ts i. (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) ts))"
      let ?fun = "\<lambda> Cs ts i. fill_holes (Cs ! i) (?list Cs ts i)"
      let ?val = "\<lambda> Cs ts i. map (?fun Cs ts) [0..<length Cs] ! i"
      fix i
      assume i: "i < length Cs"
      hence id: "?val Cs ts i = ?fun Cs ts i" "?val Cs ss i = ?fun Cs ss i" by auto
      from i have mem: "Cs ! i \<in> set Cs" by auto
      {
        fix j
        have "(j < length (?list Cs ts i) \<longrightarrow> (?list Cs ss i ! j, ?list Cs ts i ! j) \<in> R)
          \<and> num_holes (Cs ! i) = length (?list Cs ss i)" using MFun(3,4) nh i
        proof (induct Cs arbitrary: ts ss i j)
          case (Cons C Cs)
          let ?n = "num_holes C"
          from Cons(4) have len_ss: "length ss = ?n + listsum (map num_holes Cs)" by auto
          note len_ts = len_ss[unfolded Cons(3)] (* also for ts *)
          from len_ss len_ts have take: "length (take ?n ss) = length (take ?n ts)" by auto
          from len_ss have min_ss: "min (length ss) ?n = ?n" by auto
          from len_ts have min_ts: "min (length ts) ?n = ?n" by auto
          show ?case
          proof (cases i)
            case 0
            with Cons(2)[of j] show ?thesis by (simp add: min_ss)
          next
            case (Suc k)
            obtain ss' where ss: "ss = take ?n ss @ ss'" and ss': "ss' = drop ?n ss" by (metis append_take_drop_id)
            {
              fix n i
              have "drop (?n + n) ss = drop n ss'" "ss ! (?n + i) = ss' ! i" 
                unfolding arg_cong[OF ss, of "drop (?n + n)"] 
                unfolding arg_cong[OF ss, of nth] nth_append
                unfolding drop_append ss' by (auto simp: min_ss)
            } note id_ss = this
            from len_ts obtain ts' where ts: "ts = take ?n ts @ ts'" and ts': "ts' = drop ?n ts" by (metis append_take_drop_id)
            {
              fix n i
              have "drop (?n + n) ts = drop n ts'" "ts ! (?n + i) = ts' ! i"
                unfolding arg_cong[OF ts, of "drop (?n + n)"] 
                unfolding arg_cong[OF ts, of nth] nth_append
                unfolding drop_append ts' by (auto simp: min_ts)
            } note id_ts = this
            note IH = Cons(1)
            from Cons(3) have len: "length ss' = length ts'" unfolding ss' ts' by auto
            note IH = IH[OF _ this]
            from Cons(4) have nh: "listsum (map num_holes Cs) = length ss'" unfolding ss' by auto
            note IH = IH[OF _ this]
            from Cons(5) Suc have "k < length Cs" by auto
            note IH = IH[OF _ this]
            {
               fix i
               assume "i < length ts'"
               hence "?n + i < length ts" unfolding len_ts ts' nh len by auto
               from Cons(2)[OF this, unfolded id_ss id_ts]
               have "(ss' ! i, ts' ! i) \<in> R" .
            }
            note IH = IH[OF this, rule_format]
            let ?k = "listsum (map num_holes (take k Cs))"
            have "?thesis = ((j < length ts' - ?k \<and> j < num_holes (Cs ! k) \<longrightarrow>
                (drop ?k ss' ! j, drop ?k ts' ! j) \<in> R) \<and>
                num_holes (Cs ! k) = min (length ss' - ?k) (num_holes (Cs ! k)))" unfolding Suc by (simp add: id_ss id_ts)
            also have "..."
            proof (intro allI conjI impI, elim conjE)
              have "num_holes (Cs ! k) = length (take (num_holes (Cs ! k)) (drop ?k ss'))" using IH by blast
              also have "\<dots> = min (length ss' - ?k) (num_holes (Cs ! k))" by simp
              finally show "num_holes (Cs ! k) = min (length ss' - ?k) (num_holes (Cs ! k))" .
            next
              assume "j < length ts' - ?k" "j < num_holes (Cs ! k)"
              thus "(drop ?k ss' ! j, drop ?k ts' ! j) \<in> R"
                using IH[THEN conjunct1, THEN mp, of j] by auto
            qed
            finally show ?thesis .
          qed 
        qed auto
      }
      note main = this
      have CiF: "funas_mctxt (Cs ! i) \<subseteq> F" using F(2) i unfolding set_conv_nth by auto
      {
        have id: "?fun Cs ts i =f (Cs ! i, ?list Cs ts i)"
          by (rule eqfI, insert main MFun(4), auto)
        have "UNION (set (?list Cs ts i)) funas_term \<subseteq> UNION (set ts) funas_term" 
          by (rule UN_mono[OF subset_trans[OF set_take_subset set_drop_subset]], auto)
        hence ts_list: "UNION (set (?list Cs ts i)) funas_term \<subseteq> F" using F by auto
        have "funas_term (?fun Cs ts i) \<subseteq> funas_mctxt (Cs ! i) \<union> UNION (set (?list Cs ts i)) funas_term"
          unfolding eqf_funas_term[OF id] by auto
        also have "funas_mctxt (Cs ! i) \<subseteq> F" by (rule CiF)
        also have "UNION (set (?list Cs ts i)) funas_term \<subseteq> F" by fact
        finally have ts: "funas_term (?fun Cs ts i) \<subseteq> F" by blast
        have id: "?fun Cs ss i =f (Cs ! i, ?list Cs ss i)"
          by (rule eqfI, insert main MFun(4), auto)
        have "funas_term (?fun Cs ss i) \<subseteq> funas_mctxt (Cs ! i) \<union> UNION (set (?list Cs ss i)) funas_term"
          unfolding eqf_funas_term[OF id] by auto
        also have "funas_mctxt (Cs ! i) \<subseteq> F" by (rule CiF)
        also have "UNION (set (?list Cs ss i)) funas_term \<subseteq> UNION (set ss) funas_term" 
          by (rule UN_mono[OF subset_trans[OF set_take_subset set_drop_subset]], auto)
        also have "\<dots> \<subseteq> F" using F by auto
        finally have ss: "funas_term (?fun Cs ss i) \<subseteq> F" by blast
        note ts ss ts_list
      } note F_main = this
      have "((?val Cs ss i, ?val Cs ts i) \<in> R) = ((?fun Cs ss i, ?fun Cs ts i) \<in> R)" unfolding id ..
      also have "..." 
        by (rule MFun(1)[OF mem eqfI[OF refl] _ _ F_main(2-3)], insert main MFun(4), auto)
      finally show "(?val Cs ss i, ?val Cs ts i) \<in> R" .
      show "funas_term (?val Cs ss i) \<subseteq> F" unfolding id by fact
      show "funas_term (?val Cs ts i) \<subseteq> F" unfolding id by fact
    qed 
  qed
  show ?thesis
    by (rule conjI[OF R eqfI], insert eqfE[OF ass(1)] ass(3), auto)
qed

fun map_mctxt :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)mctxt \<Rightarrow> ('g,'v)mctxt" where
  "map_mctxt _ (MVar x) = (MVar x)"
| "map_mctxt _ (MHole) = MHole"
| "map_mctxt fg (MFun f Cs) = MFun (fg f) (map (map_mctxt fg) Cs)"

fun ground_mctxt :: "('f,'v)mctxt \<Rightarrow> bool" where 
  "ground_mctxt (MVar _) = False"
| "ground_mctxt MHole = True"
| "ground_mctxt (MFun f Cs) = Ball (set Cs) ground_mctxt"

lemma ground_eq_fill: "t =f (C,ss) \<Longrightarrow> ground t = (ground_mctxt C \<and> (\<forall> s \<in> set ss. ground s))" 
proof (induct C arbitrary: t ss)
  case (MVar x)
  from eqfE[OF this] show ?case by simp
next
  case (MHole t ss)
  from eqfE[OF this] show ?case by (cases ss, auto)
next
  case (MFun f Cs s ss)
  from eqf_MFunE[OF MFun(2)] obtain ts sss where s: "s = Fun f ts" and len: "length ts = length Cs" "length sss = length Cs" 
    and IH: "\<And> i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)" and ss: "ss = concat sss" by metis
  {
    fix i
    assume i: "i < length Cs"
    hence "Cs ! i \<in> set Cs" by simp
    from MFun(1)[OF this IH[OF i]]
    have "ground (ts ! i) = (ground_mctxt (Cs ! i) \<and> (\<forall>a\<in>set (sss ! i). ground a))" .
  } note IH = this
  note conv = set_conv_nth
  have "?case = ((\<forall>x\<in>set ts. ground x) = ((\<forall>x\<in>set Cs. ground_mctxt x) \<and> (\<forall>a\<in>set sss. \<forall>x\<in>set a. ground x)))"
    unfolding s ss by simp
  also have "..." unfolding conv[of ts] conv[of Cs] conv[of sss] len using IH by auto
  finally show ?case by simp
qed

lemma ground_fill_holes: assumes nh: "num_holes C = length ss"
  shows "ground (fill_holes C ss) = (ground_mctxt C \<and> (\<forall> s \<in> set ss. ground s))"
  by (rule ground_eq_fill[OF eqfI[OF refl nh]])

lemma split_vars_ground: "split_vars t = (C,xs) \<Longrightarrow> ground_mctxt C"
proof (induct t arbitrary: C xs)
  case (Fun f ts C xs)
  from Fun(2)[simplified] obtain Cs where C: "C = MFun f Cs" and Cs: "Cs = map (fst \<circ> split_vars) ts" by auto
  show ?case unfolding C ground_mctxt.simps
  proof
    fix C
    assume "C \<in> set Cs"
    from this[unfolded Cs] obtain t where t: "t \<in> set ts" and C: "C = fst (split_vars t)" unfolding o_def by auto
    from C obtain xs where split: "split_vars t = (C,xs)" by (cases "split_vars t", auto)
    show "ground_mctxt C"
      by (rule Fun(1)[OF t split]) 
  qed
qed auto

lemma split_vars_ground_vars: assumes "ground_mctxt C" and "num_holes C = length xs" 
  shows "split_vars (fill_holes C (map Var xs)) = (C, xs)"
  using assms
proof (induct C arbitrary: xs)
  case (MHole xs)
  thus ?case by (cases xs, auto)
next
  case (MFun f Cs xs)
  have "fill_holes (MFun f Cs) (map Var xs) =f (MFun f Cs, map Var xs)"
    by (rule eqfI, insert MFun(3), auto)
  from eqf_MFunE[OF this] 
  obtain ts xss where fh: "fill_holes (MFun f Cs) (map Var xs) = Fun f ts"
    and lent: "length ts = length Cs"
    and lenx: "length xss = length Cs"
    and args: "\<And>i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, xss ! i)"
    and id: "map Var xs = concat xss" by auto
  from arg_cong[OF id, of "map the_Var"] have id2: "xs = concat (map (map the_Var) xss)" 
    by (auto simp: o_def map_concat)
  {
    fix i
    assume i: "i < length Cs"
    hence mem: "Cs ! i \<in> set Cs" by auto
    with MFun(2) have ground: "ground_mctxt (Cs ! i)" by auto
    have "map Var (map the_Var (xss ! i)) = map id (xss ! i)" unfolding map_map o_def map_eq_conv
    proof
      fix x
      assume "x \<in> set (xss ! i)"
      with lenx i have "x \<in> set (concat xss)" by auto
      from this[unfolded id[symmetric]] show "Var (the_Var x) = id x" by auto
    qed
    hence idxss: "map Var (map the_Var (xss ! i)) = xss ! i" by auto
    note rec = eqfE[OF args[OF i]]
    note IH = MFun(1)[OF mem ground, of "map the_Var (xss ! i)", unfolded rec(2) idxss rec(1)[symmetric]]
    from IH have "split_vars (ts ! i) = (Cs ! i, map the_Var (xss ! i))" by auto
    note this idxss
  }
  note IH = this
  have "?case = (map fst (map split_vars ts) = Cs \<and> concat (map snd (map split_vars ts)) = concat (map (map the_Var) xss))"
    unfolding fh unfolding id2 by auto
  also have "\<dots>"
  proof (rule conjI[OF nth_equalityI[OF _ allI[OF impI]] arg_cong[of _ _ concat, OF nth_equalityI, rule_format]], unfold length_map lent lenx)
    fix i
    assume i: "i < length Cs" 
    with arg_cong[OF IH(2)[OF this], of "map the_Var"]
      IH[OF this] show "map snd (map split_vars ts) ! i = map (map the_Var) xss ! i" using lent lenx by auto
  qed (insert IH lent, auto)
  finally show ?case .
qed auto

lemma ground_map_mctxt[simp]: "ground_mctxt (map_mctxt fg C) = ground_mctxt C"
  by (induct C, auto)

lemma num_holes_map_mctxt[simp]: "num_holes (map_mctxt fg C) = num_holes C"
proof (induct C)
  case (MFun f Cs)
  thus ?case by (induct Cs, auto)
qed auto

lemma split_vars_map_mctxt: assumes split: "split_vars t = (map_mctxt fg C, xs)"
  shows "split_vars (fill_holes C (map Var xs)) = (C, xs)"
proof -
  from split_vars_ground[OF split] have ground: "ground_mctxt C" by simp
  from split_vars_num_holes[of t, unfolded split] have nh: "num_holes C = length xs" by auto
  show ?thesis
    by (rule split_vars_ground_vars[OF ground nh])
qed

lemma subst_eq_map_decomp: assumes "t \<cdot> \<sigma> = map_term fg s"
  shows "\<exists> C xs \<delta>s. s =f (C,\<delta>s) \<and> split_vars t = (map_mctxt fg C, xs) \<and> (\<forall> i < length xs. 
    \<sigma> (xs ! i) = map_term fg (\<delta>s ! i))" 
  using assms
proof (induct t arbitrary: s)
  case (Var x s)
  show ?case
    by (rule exI[of _ MHole], rule exI[of _ "[x]"], rule exI[of _ "[s]"], insert Var, auto)
next
  case (Fun g ts s)
  from Fun(2) obtain f ss where s: "s = Fun f ss" and g: "g = fg f" by (cases s, auto)
  from Fun(2)[unfolded s] have id: "map (\<lambda> t. t \<cdot> \<sigma>) ts = map (map_term fg) ss" by auto
  from arg_cong[OF this, of length] have len: "length ts = length ss" by auto
  from map_nth_conv[OF id] have args: "\<And> i. i < length ts \<Longrightarrow> ts ! i \<cdot> \<sigma> = map_term fg (ss ! i)" by auto
  let ?P = "\<lambda> C xs \<delta>s i. ss ! i =f (C, \<delta>s) \<and>
          split_vars (ts ! i) = (map_mctxt fg C, xs) \<and>
          (\<forall>i<length xs. \<sigma> (xs ! i) = map_term fg (\<delta>s ! i))"
  {
    fix i
    assume i: "i < length ts"
    hence mem: "ts ! i \<in> set ts" by auto
    note IH = Fun(1)[OF this args[OF i]]
  }
  hence "\<forall> i. \<exists> C xs \<delta>s. i < length ts \<longrightarrow> ?P C xs \<delta>s i" by blast
  from choice[OF this] obtain Cs where "\<forall> i. \<exists> xs \<delta>s. i < length ts \<longrightarrow> ?P (Cs i) xs \<delta>s i" by blast
  from choice[OF this] obtain xss where "\<forall> i. \<exists> \<delta>s. i < length ts \<longrightarrow> ?P (Cs i) (xss i) \<delta>s i" by blast
  from choice[OF this] obtain \<delta>ss where IH: "\<And> i. i < length ts \<Longrightarrow> ?P (Cs i) (xss i) (\<delta>ss i) i" by blast
  let ?n = "[0 ..< length ts]"
  let ?Cs = "map Cs ?n"
  let ?C = "MFun f ?Cs"
  let ?xs = "concat (map xss ?n)"
  let ?\<delta>s = "concat (map \<delta>ss ?n)"
  let ?g = "fg f"
  show ?case unfolding s g 
  proof (rule exI[of _ ?C], rule exI[of _ ?xs], rule exI[of _ ?\<delta>s], intro conjI)
    show "Fun f ss =f (?C, ?\<delta>s)" 
      by (rule eqf_MFunI, insert IH len, auto)
  next
    have 
      "(split_vars (Fun ?g ts) = (map_mctxt fg ?C, ?xs)) 
        = (map (fst \<circ> split_vars) ts = map (map_mctxt fg \<circ> Cs) [0..<length ss] 
          \<and>concat (map (snd \<circ> split_vars) ts) = ?xs)" 
      (is "?goal = _")
      using len by auto
    also have "\<dots>"
      by (rule conjI[OF nth_map_conv arg_cong[of _ _ concat, OF nth_equalityI]], insert IH len, auto)
    finally show ?goal .
  next
    show "\<forall> i < length ?xs. \<sigma> (?xs ! i) = map_term fg (?\<delta>s ! i)"
    proof (rule concat_all_nth, simp, unfold length_map length_upt)
      fix i
      assume "i < length ts - 0"
      hence i: "i < length ts" by auto
      from IH[OF i] have "split_vars (ts ! i) = (map_mctxt fg (Cs i), xss i)" by blast
      from split_vars_map_mctxt[OF this] split_vars_num_holes[of "fill_holes (Cs i) (map Var (xss i))"]
      have len: "length (xss i) = num_holes (Cs i)" by simp
      also have "\<dots> = length (\<delta>ss i)" by (rule eqfE(2), insert IH[OF i], auto)
      finally 
      show "length (map xss [0..<length ts] ! i) = length (map \<delta>ss [0..<length ts] ! i)" using i by auto
    qed (insert IH, auto)
  qed
qed

lemma map_term_fill_holes: "num_holes C = length ss \<Longrightarrow> map_term fg (fill_holes C ss) =f (map_mctxt fg C, map (map_term fg) ss)"
proof (induct C arbitrary: ss)
  case (MHole ss)
  thus ?case by (cases ss, auto)
next
  case MVar thus ?case by auto
next
  case (MFun f Cs ss)
  from MFun(2) have "fill_holes (MFun f Cs) ss =f (MFun f Cs, ss)" by auto
  from eqf_MFunE[OF this] obtain ts sss where fh: "fill_holes (MFun f Cs) ss = Fun f ts"
    and lts: "length ts = length Cs"
    and lsss: "length sss = length Cs"
    and args: "\<And>i. i < length Cs \<Longrightarrow> ts ! i =f (Cs ! i, sss ! i)"
    and sss: "ss = concat sss" by auto
  {
    fix i
    assume i: "i < length Cs"
    hence mem: "Cs ! i \<in> set Cs" by auto
    from MFun(1)[OF mem]  eqfE[OF args[OF i]] have 
      "map_term fg (ts ! i) =f (map_mctxt fg (Cs ! i), map (map_term fg) (sss ! i))" by auto
  } note IH = this
  show ?case unfolding fh 
    unfolding map_mctxt.simps sss map_concat map_term.simps
  proof (rule eqf_MFunI, unfold length_map, simp add: lsss, simp add: lts)
    fix i
    assume i: "i < length Cs"
    have "(map (map_term fg) ts ! i =f (map (map_mctxt fg) Cs ! i, map (map (map_term fg)) sss ! i)) =
      (map_term fg (ts ! i) =f (map_mctxt fg (Cs ! i), map (map_term fg) (sss ! i)))" (is "?goal = _")
      using i lts lsss by auto
    also have "\<dots>" by (rule IH[OF i])
    finally show ?goal .
  qed
qed

fun poss_mctxt :: "('f,'v)mctxt \<Rightarrow> pos set" where
  "poss_mctxt (MVar x) = {\<epsilon>}"
| "poss_mctxt MHole = {}"
| "poss_mctxt (MFun f cs) = {\<epsilon>} \<union> \<Union>(set (map (\<lambda> i. (\<lambda> p. i <# p) ` poss_mctxt (cs ! i)) [0 ..< length cs]))"

lemma poss_mctxt_simp[simp]: "poss_mctxt (MFun f cs) = {\<epsilon>} \<union> {i <# p | i p. i < length cs \<and> p \<in> poss_mctxt (cs ! i)}"
  by auto
declare poss_mctxt.simps(3)[simp del]

fun hole_poss :: "('f,'v)mctxt \<Rightarrow> pos set" where
  "hole_poss (MVar x) = {}"
| "hole_poss MHole = {\<epsilon>}"
| "hole_poss (MFun f cs) = \<Union>(set (map (\<lambda> i. (\<lambda> p. i <# p) ` hole_poss (cs ! i)) [0 ..< length cs]))"

lemma hole_poss_simp[simp]: "hole_poss (MFun f cs) = {i <# p | i p. i < length cs \<and> p \<in> hole_poss (cs ! i)}"
  by auto
declare hole_poss.simps(3)[simp del]


fun
  term_to_mctxt :: "('f, 'v) term \<Rightarrow> ('f, 'v) mctxt" where 
   "term_to_mctxt (Var x) = (MVar x)"
  | "term_to_mctxt (Fun f ts) = MFun f [ term_to_mctxt t . t \<leftarrow> ts]"

lemma term_to_mctxt_numholes [simp]:
  "num_holes (term_to_mctxt t) = 0"
proof (induct t, auto) qed

lemma term_to_mctxt_fill_holes [simp]:
  "fill_holes (term_to_mctxt t) [] = t"
  proof (induct t)
    case (Fun f ts) 
    then have "fill_holes (term_to_mctxt (Fun f ts)) [] = Fun f (map (\<lambda>i. (ts!i)) [0..<length ts])"
      unfolding term_to_mctxt.simps partition_holes_fill_holes_conv partition_holes_nil map_map  by auto
    also have "... = Fun f ts" using map_nth by auto
    ultimately show ?case by auto
  qed (auto)

lemma term_to_mctxt_sound:
  "t =f (term_to_mctxt t, [])"
proof (rule eqfI, insert term_to_mctxt_numholes term_to_mctxt_fill_holes[symmetric] list.size(3)[symmetric], auto)
qed

fun
  subst_apply_mctxt :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v, 'w) gsubst \<Rightarrow> ('f, 'w) mctxt"
  (infixl "\<cdot>mc" 67) where
  "MHole \<cdot>mc _ = MHole"
|  "(MVar x) \<cdot>mc \<sigma> = term_to_mctxt (\<sigma> x)"
| "(MFun f cs) \<cdot>mc \<sigma> = MFun f [c \<cdot>mc \<sigma> . c \<leftarrow> cs]"


lemma subst_apply_mctxt_numholes: 
  shows "num_holes (c \<cdot>mc \<sigma>) = num_holes c"
  proof (induct c arbitrary: \<sigma>) 
    case (MFun f cs)
    have "num_holes (MFun f cs \<cdot>mc \<sigma>) = listsum [num_holes (c \<cdot>mc \<sigma>) . c \<leftarrow> cs]"
      unfolding subst_apply_mctxt.simps num_holes.simps map_map comp_def by auto
    also have "... = listsum [num_holes c . c \<leftarrow> cs]" using MFun(1) 
      by (metis (lifting, no_types) map_cong)
    ultimately show ?case by auto
  qed (insert term_to_mctxt_numholes, auto)

lemma subst_apply_mctxt_fill_holes: 
  assumes nh: "num_holes c = length ts"
  shows "(fill_holes c ts) \<cdot> \<sigma> = fill_holes (c \<cdot>mc \<sigma>) [ ti \<cdot> \<sigma> . ti \<leftarrow> ts]"
  using nh proof (induct c arbitrary: ts)
  case MHole 
    then obtain t where ts: "ts = [t]" 
      unfolding num_holes.simps unfolding One_nat_def using Suc_length_conv length_0_conv by metis
    show ?case unfolding ts by simp
  next 
  case (MVar x)
  then have ts: "ts = []" using length_0_conv by auto 
  show ?case unfolding ts by auto
  next
  case (MFun f cs) 
  note IH = MFun(1)
  note nh = MFun(2)[unfolded num_holes.simps]
  let ?c = "MFun f cs"
  let ?cs\<sigma> = "map (\<lambda>c. c \<cdot>mc \<sigma>) cs"

  {
    fix i
    assume i: "i < length cs"
    have nh_map: "\<And> j. j < length cs \<Longrightarrow> num_holes (cs!j) = num_holes (?cs\<sigma> ! j)" 
      using nth_map subst_apply_mctxt_numholes by metis

    have "fill_holes (cs ! i) (partition_holes ts cs ! i) \<cdot> \<sigma> = 
      fill_holes ((cs ! i) \<cdot>mc \<sigma>) (partition_holes [ ti \<cdot> \<sigma> . ti \<leftarrow> ts] cs ! i)"
      using IH[OF nth_mem[OF i] partition_holes_nth_length[OF nh i], unfolded partition_holes_map_nth[OF i, symmetric]]  by auto
    also have "... = fill_holes (?cs\<sigma> ! i) (partition_holes [ti \<cdot> \<sigma> . ti \<leftarrow> ts] ?cs\<sigma> ! i)"
      unfolding nth_map[OF i] using partition_holes_map_ctxt[OF _ nh_map] length_map by metis
    ultimately have "fill_holes (cs ! i) (partition_holes ts cs ! i) \<cdot> \<sigma> = fill_holes (?cs\<sigma> ! i) (partition_holes [ti \<cdot> \<sigma> . ti \<leftarrow> ts] ?cs\<sigma> ! i)"
      by auto
    } note ith = this

  have "fill_holes ?c ts \<cdot> \<sigma> = Fun f [fill_holes (?cs\<sigma> ! i) (partition_holes [ ti \<cdot> \<sigma> . ti \<leftarrow> ts] ?cs\<sigma> ! i) . i \<leftarrow> [0..<length cs]]"  
  unfolding partition_holes_fill_holes_conv map_map using ith using comp_def  by auto 
  also have "... = fill_holes (?c \<cdot>mc \<sigma>) [ ti \<cdot> \<sigma> . ti \<leftarrow> ts]"  
    unfolding subst_apply_mctxt.simps partition_holes_fill_holes_conv length_map ..
  ultimately show ?case by auto
qed
  
lemma subst_apply_mctxt_sound: 
  assumes "t =f (c,ts)"
  shows "t \<cdot> \<sigma> =f (c \<cdot>mc \<sigma>, [ ti \<cdot> \<sigma> . ti \<leftarrow> ts])"
proof (rule eqfI, insert subst_apply_mctxt_numholes subst_apply_mctxt_fill_holes[OF eqfE(2)[OF assms]] eqfE[OF assms] eqfE(2)[OF assms, symmetric], auto) qed


fun fill_holes_mctxt :: "('f,'v)mctxt \<Rightarrow> ('f,'v)mctxt list \<Rightarrow> ('f,'v)mctxt" where
  "fill_holes_mctxt (MVar x) _ = MVar x"
| "fill_holes_mctxt MHole [t] = t"
| "fill_holes_mctxt (MFun f cs) ts = (MFun f (map (\<lambda> i. fill_holes_mctxt (cs ! i) 
    (take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ts))) [0 ..< length cs]))"

lemma partition_holes_fill_holes_mctxt_conv:
  shows "fill_holes_mctxt (MFun f cs) ts = MFun f [ fill_holes_mctxt (cs!i) (partition_holes ts cs!i) . i \<leftarrow> [0..<length cs]]"
  unfolding partition_holes_def by auto

lemma fill_holes_mctxt_MFun: assumes lCs: "length Cs = length ts"
  and lss: "length ss = length ts"
  and rec: "\<And> i. i < length ts \<Longrightarrow> num_holes (Cs ! i) = length (ss ! i) \<and> fill_holes_mctxt (Cs ! i) (ss ! i) = ts ! i"
  shows "fill_holes_mctxt (MFun f Cs) (concat ss) = MFun f ts" 
  unfolding fill_holes_mctxt.simps mctxt.simps
    by (rule conjI[OF refl], rule fill_holes_arbitrary[OF lCs lss rec])

lemma fill_holes_mctxt_numholes: 
  assumes len_ds: "length ds = num_holes c"
  shows "num_holes (fill_holes_mctxt c ds) = listsum [num_holes di . di \<leftarrow> ds]"
using len_ds proof (induct c arbitrary: ds)
  case (MVar x) then have ds: "ds = []" by auto
  show ?case unfolding ds by simp
next
  case (MHole) then
  obtain d where ds: "ds = [d]"  unfolding num_holes.simps unfolding One_nat_def using Suc_length_conv length_0_conv by metis
  show ?case unfolding ds by simp
next 
  case (MFun f cs)
  note IH=MFun(1)
  let ?c = "MFun f cs"

  from MFun(2) have l_ds: "listsum (map num_holes cs) = length ds" by simp

  have "map num_holes [(fill_holes_mctxt (cs ! i) (partition_holes ds cs ! i)) . i \<leftarrow> [0..<length cs]] 
         = [ listsum (map num_holes (partition_holes ds cs ! i)) . i \<leftarrow> [0..<length cs]]"
  proof (induct "length cs", insert IH[OF _ partition_holes_nth_length[OF l_ds, symmetric]], auto) qed
  also have "... = map listsum (map (map num_holes) (map (\<lambda> i. partition_holes ds cs ! i) [0..<length cs]))" by auto
  also have "... = map listsum (map (map num_holes) (partition_holes ds cs))" (is "_ = ?R")
    unfolding partition_holes_len[of ds cs, symmetric] map_nth ..
  ultimately have "num_holes (fill_holes_mctxt ?c ds) = listsum ?R"
    unfolding partition_holes_fill_holes_mctxt_conv num_holes.simps by metis
  also have "... = listsum (map num_holes ds)"
      unfolding listsum_concat[symmetric] map_concat[symmetric] concat_partition_holes[OF l_ds] by auto
    ultimately show ?case by auto
qed

lemma fill_holes_mctxt_fill_holes: 
  assumes len_ds: "length ds = num_holes c"
  and nh: "num_holes (fill_holes_mctxt c ds) = length ss"
  shows "fill_holes (fill_holes_mctxt c ds) ss = fill_holes c [fill_holes (ds!i) (partition_holes ss ds!i) . i \<leftarrow> [0..< num_holes c ]]"
using nh len_ds proof (induct c arbitrary: ss ds)
  case (MVar x) then have ds: "ds = []"  by auto
  show ?case unfolding ds by simp
next
  case (MHole) then
  obtain d where ds: "ds = [d]" 
    unfolding num_holes.simps
    using Suc_length_conv length_0_conv One_nat_def by metis
  have "length ss = num_holes d"
    unfolding MHole(1)[symmetric] ds by simp
  then show ?case unfolding ds partition_holes_step by auto
next 
  case (MFun f cs)
  note IH=MFun(1)
  let ?c = "MFun f cs"

  def csi \<equiv> "\<lambda> j . fill_holes_mctxt (cs ! j) (partition_holes ds cs ! j)"
  def ssi \<equiv> "\<lambda> j . partition_holes ss (map csi [0..<length cs]) ! j"

  {
    fix i 
    assume i: "i < length cs"
    hence i': "i < length [0..<length cs]" by auto
    hence i'': "[0..<length cs] ! i = i" by auto
    have i_ph: "i < length (partition_holes ds cs)" using i partition_holes_len by metis

    let ?dsis = "partition_holes ds cs ! i"
    let ?td_l = "\<lambda> ls . take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ls)"    
    let ?td_nh = "\<lambda> ls . take (length ?dsis) (drop (listsum (map length (take i (partition_holes ds cs)))) ls)"

    have l_dsi: "\<And>  j. j < length cs \<Longrightarrow> length (partition_holes ds cs ! j) = num_holes (cs!j)" 
      by (metis MFun.prems(2) num_holes.simps(3) partition_holes_nth_length)
    have l_nh_conv: "map length (partition_holes ds cs) = map num_holes cs"
      by (metis l_dsi length_map map_nth_eq_conv partition_holes_len)

    { 
      fix ls
      have "?td_l ls = ?td_nh ls" unfolding l_dsi[OF i] unfolding take_map[symmetric] l_nh_conv by auto
    } note td_l_nh_conv = this
      
    have ph_ds_i: "partition_holes ds cs ! i = ?td_l ds" unfolding partition_holes_take_drop[OF i] ..
    have ph_ss_i: "partition_holes (ssi i) (partition_holes ds cs ! i) = ?td_l (partition_holes ss ds)" 
    proof -
      from i have tk: "take i [0..<length cs] = [0..<i]" using take_upt by auto
      have nh_csi: "\<And>  j. j < length cs \<Longrightarrow> num_holes (csi j) = listsum [num_holes di . di \<leftarrow> partition_holes ds cs ! j]"
        unfolding csi_def using fill_holes_mctxt_numholes[OF l_dsi] i  by auto
      have nh_csjs: "[num_holes (csi j) . j \<leftarrow> [0..<i]] = [listsum [num_holes di . di \<leftarrow> partition_holes ds cs ! j] . j \<leftarrow> [0..<i]]"
        using nh_csi i by auto 
      have i_csis: "i < length (map csi [0..<length cs])" unfolding length_map using i by auto
      have ssi_take_drop: 
        "ssi i = take (listsum [num_holes di . di \<leftarrow> partition_holes ds cs ! i]) 
        (drop (listsum [listsum [num_holes di . di \<leftarrow> partition_holes ds cs ! j] . j \<leftarrow> [0..<i]]) ss)"
        unfolding ssi_def partition_holes_take_drop[OF i_csis] take_map tk map_map o_def nth_map[OF i'] i'' nh_csi[OF i] nh_csjs  ..
      
      {
        fix j
        assume j: "j < length (partition_holes ds cs ! i)"

        let ?a = "map num_holes ?dsis"
        let ?l = "map num_holes (take j ?dsis)"
        let ?r = "map num_holes (drop j ?dsis)"


        have "listsum ?a - listsum ?l = listsum (?l @ ?r) - listsum ?l"
          using append_take_drop_id[symmetric] take_map drop_map by metis
        then have ls_r: "listsum ?a - listsum ?l = listsum  ?r" by auto

        from j obtain di dis where "drop j ?dsis = di # dis"
          by (metis drop_Suc_conv_tl)
        moreover have "?dsis ! j = drop j ?dsis ! 0"
          using j nth_drop by auto
        ultimately have "num_holes (?dsis ! j) \<le> listsum ?r" proof (cases "?r", auto) qed
        then have "min (num_holes (?dsis ! j)) (listsum ?a - listsum ?l) = num_holes (?dsis ! j)" using ls_r by auto
      } note takes = this

      have t_i_suci_conv: "take i (map num_holes cs) @ (num_holes (cs ! i) # drop (Suc i) (map num_holes cs)) = map num_holes cs"
        by (metis length_map nth_map id_take_nth_drop i)
      have "listsum (map length (take i (partition_holes ds cs))) + length (partition_holes ds cs ! i) 
            \<le> listsum (take i (map num_holes cs) @ (num_holes (cs ! i) # drop (Suc i) (map num_holes cs)))"
        unfolding take_map[symmetric] l_nh_conv l_dsi[OF i] listsum_append listsum_simps(2) by auto
      also have "... = length ds"
        unfolding MFun(3) t_i_suci_conv by simp
      ultimately have ti_le_len_ds: "listsum (map length (take i (partition_holes ds cs))) + length (partition_holes ds cs ! i) \<le> length ds"
        by auto
      then have 
        "?td_nh [0..<length ds] 
         = [(listsum (map length (take i (partition_holes ds cs)))) ..< (listsum (map length (take i (partition_holes ds cs)))) + (length ?dsis)]"
        unfolding drop_upt using take_upt by auto
      also have "... = [ id (listsum (map length (take i (partition_holes ds cs))) + j). j \<leftarrow> [0..<length ?dsis]]" 
        unfolding map_upt_add'[symmetric] by auto
      ultimately have td_nh_ixs: 
        "?td_nh [0..<length ds] = [ j + listsum (map length (take i (partition_holes ds cs))). j \<leftarrow> [0..<length ?dsis]]" by auto

      {
        fix j
        assume j: "j < length ?dsis"
        then have b: "(listsum (take i (map length (partition_holes ds cs)))) + j \<le> length ds" 
          unfolding take_map using ti_le_len_ds by auto

        have "ds ! (j + listsum (map length (take i (partition_holes ds cs)))) = ?dsis ! j" 
          unfolding partition_holes_take_drop[OF i] 
            unfolding l_dsi[OF i, symmetric] nth_take[OF j] take_map[symmetric] l_nh_conv[symmetric]
            unfolding nth_drop[OF b, symmetric, unfolded nat_add_commute] ..
      } note dsis_j_conv = this

      {
        fix j
        assume j: "j < length ?dsis"
        then have min: "min j (num_holes (cs!i)) = j" 
          by (metis l_dsi[OF i] less_not_sym min_less_iff_disj nat_less_cases)
        let ?s = "listsum (map length (take i (partition_holes ds cs)))"

        have dsi_j: "take j (drop ?s ds) = take j ?dsis"
          unfolding partition_holes_take_drop[OF i] take_take min take_map[symmetric] l_nh_conv ..
        
        have map_l_ph_conv: "map length (take i (partition_holes ds cs)) = map num_holes (take i cs)" 
          unfolding take_map[symmetric] l_nh_conv ..

        from i have i_leq: "i \<le> length cs" by auto
        have dsjs: 
          "take ?s ds = concat [ partition_holes ds cs ! j. j \<leftarrow> [0..<i]]"
          unfolding map_l_ph_conv concat_partition_holes_upto[OF i_leq] 
          unfolding take_upt_idx[OF i] map_map o_def ..

        have conv:
          "take (j + ?s) ds = concat [ partition_holes ds cs ! j. j \<leftarrow> [0..<i]] @ take j ?dsis" 
          using take_add nat_add_commute dsi_j dsjs by metis
        
        have 
          "listsum (map num_holes (take (j + listsum (map length (take i (partition_holes ds cs)))) ds))
          = listsum (map num_holes (take j ?dsis))
          + (\<Sum>k\<leftarrow>[0..<i]. listsum (map num_holes (partition_holes ds cs ! k)))"
          unfolding conv map_append listsum_append map_concat listsum_concat map_map o_def
          unfolding map_map
          using nat_add_commute by auto
      } note dsis_j_conv2 = this
      
      have "?td_l (partition_holes ss ds)
        = [take (num_holes (ds ! i)) (drop (listsum (map num_holes (take i ds))) ss) . i \<leftarrow> ?td_nh [0..<length ds]]"
        unfolding td_l_nh_conv partition_holes_def drop_map take_map ..
      also have 
        "... 
        = [ take (num_holes (?dsis ! j)) 
        (drop (listsum (map num_holes (take j ?dsis))
        + (\<Sum>k\<leftarrow>[0..<i]. listsum (map num_holes (partition_holes ds cs ! k)))) ss)
        . j \<leftarrow> [0..<length ?dsis]]"
        unfolding td_nh_ixs map_map o_def using dsis_j_conv dsis_j_conv2 by auto
      also have 
        "... = partition_holes (ssi i) ?dsis"
        unfolding partition_holes_def[of "(ssi i)"] 
        unfolding ssi_take_drop drop_take take_take drop_drop
        using takes by auto
      ultimately show ?thesis by auto
    qed
    
    have l1: "length (?td_l ds) = num_holes (cs ! i)" by (metis l_dsi[OF i] ph_ds_i)
    have l2: "length (?td_l (partition_holes ss ds)) = num_holes (cs ! i)" by (metis l1 partition_holes_len ph_ds_i ph_ss_i)
    from l1 l2 have l3: "length (zip (?td_l ds) (?td_l (partition_holes ss ds))) = num_holes (cs ! i)" by auto

    have nh_i: "num_holes (csi i) = length (ssi i)" 
      unfolding csi_def ssi_def 
      using partition_holes_nth_length[OF MFun.prems(1)[unfolded partition_holes_fill_holes_mctxt_conv num_holes.simps], unfolded  length_map]
      by (metis (lifting, no_types) i' i'' nth_map)

    have "[ fill_holes (partition_holes ds cs ! i ! j) (partition_holes (ssi i) (partition_holes ds cs ! i) ! j) . j \<leftarrow> [0..<num_holes (cs ! i)] ]
         = map (\<lambda> j . fill_holes (?td_l ds ! j) (?td_l (partition_holes ss ds) ! j)) [0..<num_holes (cs ! i)]"
      unfolding ph_ss_i using ph_ds_i by auto
    also have 
      "... = map (\<lambda> j . (\<lambda> e. fill_holes (fst e) (snd e)) (?td_l ds ! j, ?td_l (partition_holes ss ds) ! j)) [0..<num_holes (cs ! i)]"
      by auto
    also have 
      "... = map (\<lambda> j . (\<lambda> e. fill_holes (fst e) (snd e)) (zip (?td_l ds) (?td_l (partition_holes ss ds)) ! j)) [0..<num_holes (cs ! i)]" 
      using nth_zip[symmetric] l1 l2 proof (induct "[0..<num_holes (cs ! i)]", auto) qed
    also have 
      "... = map (\<lambda> j . map (\<lambda> e. fill_holes (fst e) (snd e)) (zip (?td_l ds) (?td_l (partition_holes ss ds))) ! j) [0..<num_holes (cs ! i)]" 
      using l3 nth_map[symmetric] proof (induct "[0..<num_holes (cs ! i)]", auto) qed
    also have 
      "... = map (\<lambda> j . map (\<lambda> e. fill_holes (fst e) (snd e)) (zip (?td_l ds) (?td_l (partition_holes ss ds))) ! j) 
                 [0..<length (map (\<lambda> e. fill_holes (fst e) (snd e)) (zip (?td_l ds) (?td_l (partition_holes ss ds))))]"
      unfolding length_map using l3 by auto
    also have 
      "... = map (\<lambda> e. fill_holes (fst e) (snd e)) (zip (?td_l ds) (?td_l (partition_holes ss ds)))" 
      using map_nth by metis 
    also have 
      "... = map (\<lambda> e. fill_holes (fst e) (snd e)) (?td_l (zip ds (partition_holes ss ds)))" 
      unfolding take_zip[symmetric] drop_zip[symmetric] ..
    also have
      "... = map (\<lambda> e. fill_holes (fst e) (snd e)) (partition_holes (zip ds (partition_holes ss ds)) cs ! i)" 
      unfolding partition_holes_take_drop[OF i] ..
    also have 
      "... = partition_holes (map (\<lambda>e. fill_holes (fst e) (snd e)) (zip ds (partition_holes ss ds))) cs ! i" 
      unfolding partition_holes_map_nth[OF i, symmetric]..
    also have 
      "... = partition_holes (map (\<lambda>k. fill_holes (ds ! k) (partition_holes ss ds ! k)) [0..<num_holes ?c]) cs ! i"
      unfolding zip_nth_conv[OF partition_holes_len[symmetric]] map_map o_def partition_holes_len MFun.prems(2) by auto
    ultimately 
    have fh: 
      "(map (\<lambda>j. fill_holes (partition_holes ds cs ! i ! j) (partition_holes (ssi i) (partition_holes ds cs ! i) ! j))
            [0..<num_holes (cs ! i)]) 
       = partition_holes (map (\<lambda>k. fill_holes (ds ! k) (partition_holes ss ds ! k)) [0..<num_holes ?c]) cs ! i" 
      by auto
    have 
      "fill_holes (map csi [0..<length cs] ! i) (ssi i) = fill_holes (csi i) (ssi i)" (is "?l = _ ") 
      unfolding csi_def ssi_def  using nth_map[OF i'] i'' by metis
    also have 
      "... = fill_holes (cs!i) 
                        (map (\<lambda>j. fill_holes (partition_holes ds cs ! i ! j) 
                                             (partition_holes (ssi i) (partition_holes ds cs ! i) ! j)) 
                             [0..<num_holes (cs ! i)])" 
      using IH[OF nth_mem[OF i] nh_i[unfolded csi_def ssi_def] l_dsi[OF i]] 
      unfolding csi_def ssi_def by auto
    also have 
      "... = fill_holes (cs!i) 
                        (partition_holes (map (\<lambda>j. fill_holes (ds ! j) (partition_holes ss ds ! j)) [0..<num_holes ?c]) cs ! i)" (is "_ = ?r") 
      using fh by auto
    ultimately have "?l = ?r" by auto
  } 
  then show ?case 
    unfolding 
      ssi_def csi_def
      partition_holes_fill_holes_mctxt_conv partition_holes_fill_holes_conv length_map  by auto
qed

lemma fill_holes_mctxt_sound:
  assumes len_ds: "length ds = num_holes c"
  and len_sss: "length sss = num_holes c"
  and len_ts: "length ts = num_holes c"
  and insts: "\<And> i. i < length ds \<Longrightarrow> ts!i =f (ds!i, sss!i)"
  shows "fill_holes c ts =f (fill_holes_mctxt c ds, concat sss)"
proof (rule eqfI)
  note l_nh_i = eqfE(2)[OF insts]

  from concat_partition_holes_id[OF _ l_nh_i] len_ds len_sss
  have concat_sss: "partition_holes (concat sss) ds = sss" by auto

  then show nh: "num_holes (fill_holes_mctxt c ds) = length (concat sss)"
    unfolding fill_holes_mctxt_numholes[OF len_ds] length_concat
    by (metis l_nh_i len_ds len_sss nth_map_conv)  

  have ts: "ts = [fill_holes (ds ! i) (partition_holes (concat sss) ds ! i) . i \<leftarrow> [0..<num_holes c]]" (is "_ = ?fhs")
  proof (rule nth_equalityI)
    show l_fhs: "length ts = length ?fhs" unfolding length_map
      by (metis diff_zero len_ts length_upt)
    {
      fix i
      assume i: "i < length ts"
      then have i': "i < length [0..<num_holes c]" 
        by (metis diff_zero len_ts length_upt)
      have "ts!i = ?fhs ! i"
        unfolding nth_map[OF i']
        using eqfE(1)[OF insts[unfolded len_ds, OF i[unfolded len_ts]]] 
        by (metis concat_sss i' len_ds len_sss map_nth nth_map)
    } 
    then show "\<forall>i<length ts. ts ! i = ?fhs ! i" by auto
  qed 
  note ts = this

  show "fill_holes c ts = fill_holes (fill_holes_mctxt c ds) (concat sss)"
    unfolding fill_holes_mctxt_fill_holes[OF len_ds nh] ts ..
qed      



fun compose_mctxt :: "('f,'v)mctxt \<Rightarrow> nat \<Rightarrow> ('f,'v)mctxt \<Rightarrow> ('f,'v)mctxt" where
  "compose_mctxt C i Ci = fill_holes_mctxt C [ if i = j then Ci else MHole . j \<leftarrow> [0..<num_holes C]]"

lemma compose_mctxt_sound:
  assumes s: "s =f (C, bef @ si # aft)"
  and si: "si =f (Ci, ts)"
  and i: "i = length bef"
  shows "s =f (compose_mctxt C i Ci, bef @ ts @ aft)"
proof -
  let ?Cs = "[ if i = j then Ci else MHole . j \<leftarrow> [0..<num_holes C]]"
  let ?ts = "bef @ si # aft"
  let ?sss = "[ [b]. b \<leftarrow> bef ] @ (ts # [ [a]. a \<leftarrow> aft])"

  have 
    l_Cs : "length ?Cs = num_holes C" and
    l_ts : "length ?ts = num_holes C" and
    l_sss : "length ?sss = num_holes C"
    unfolding length_append length_map list.size(4) using eqfE(2)[OF s] by auto

  have i_le_nh: "i < num_holes C" unfolding i eqfE(2)[OF s] length_append by (auto iff: trans_less_add1)
  have concat_sss: "concat ?sss = bef @ ts @ aft" by auto

  {
    fix j
    assume j: "j < i"
    then have j': "j < length [0..<num_holes C]" using i_le_nh length_upt by auto
    have "?sss!j = [bef!j]" by (metis append_Cons_nth_left i j length_map nth_map) 
    moreover have "?ts!j = bef!j" by (metis append_Cons_nth_left i j)
    moreover from nth_map[OF j'] j' j have "?Cs!j = MHole" by force
    ultimately have "?ts!j =f (?Cs!j,?sss!j)" using eqfI by auto
    } note j_le_i = this


  from i_le_nh have "?Cs!i = Ci" by auto
  moreover from i_le_nh have "?sss!i = ts" by (metis i length_map nth_append_length)
  moreover have "?ts!i = si" using nth_append_length i by auto
  ultimately have j_eq_i: "?ts!i =f (?Cs!i,?sss!i)" using si by auto

  {
    fix j
    assume j: "j > i" and j': "j < num_holes C"
    then have j'': "j < length [0..<num_holes C]" by auto
    
    have j''': "(j - i) - 1 < length aft" using 
      j'[unfolded eqfE(2)[OF s] length_append[of bef] list.size(4)] j 
      unfolding i by auto

    from nth_append[of "[ [b]. b \<leftarrow> bef ]" _ j, unfolded length_map[of _ bef] i[symmetric]] 
    have "?sss!j = (ts # [ [a]. a \<leftarrow> aft]) ! (j - i)"  using j by auto
    moreover have "... = [ [a]. a \<leftarrow> aft] ! ((j - i) - 1)" using nth_Cons_pos j by simp
    moreover have "... = [aft ! ((j - i) - 1)]" using j''' length_map nth_map by auto
    ultimately have sssj: "?sss!j = [aft ! ((j - i) - 1)]" by auto
    
    have Csj: "?Cs!j = MHole" using nth_map[OF j''] j'' j  by force

    have "?ts!j = (si # aft) ! (j - i)" unfolding nth_append[of bef] i[symmetric] using j by simp
    moreover have "... = aft ! ((j - i) - 1)" by (metis j neq0_conv nth_Cons' zero_less_diff)
    ultimately have "?ts!j = aft ! ((j - i) - 1)" by auto
    then have "?ts!j =f (?Cs!j,?sss!j)" using sssj Csj by auto
  } note j_gr_i = this  

  from j_le_i j_eq_i j_gr_i have "\<And> j. j < length ?Cs \<Longrightarrow> ?ts!j =f (?Cs!j,?sss!j)" 
    using l_Cs linorder_neqE_nat by metis
  
  from fill_holes_mctxt_sound[OF l_Cs l_sss l_ts this, unfolded concat_sss, folded compose_mctxt.simps]
  show ?thesis unfolding eqfE(1)[OF s] by simp
qed


fun mctxt_fill_partially_mctxts :: "('f,'v) term list \<Rightarrow> ('f,'v) term list \<Rightarrow> ('f,'v) mctxt list"
  where 
    "mctxt_fill_partially_mctxts [] ts = map term_to_mctxt ts"
  | "mctxt_fill_partially_mctxts (s # ss) (t # ts) = 
        (if s = t 
         then (MHole # mctxt_fill_partially_mctxts ss ts) 
         else (term_to_mctxt t # mctxt_fill_partially_mctxts (s # ss) ts))"

fun mctxt_fill_partially_fills :: "('f,'v) term list \<Rightarrow> ('f,'v) term list \<Rightarrow> ('f,'v) term list list"
  where 
    "mctxt_fill_partially_fills [] ts = map (const []) ts"
  | "mctxt_fill_partially_fills (s # ss) (t # ts) = 
        (if (s = t) 
         then ([s] # mctxt_fill_partially_fills ss ts) 
         else ([] # mctxt_fill_partially_fills (s # ss) ts))"

lemma mctxt_fill_partially_mctxts_length [simp]:
  assumes "sublist ss ts"
  shows "length (mctxt_fill_partially_mctxts ss ts) = length ts"
  using assms proof (induct rule: sublist_induct2, insert length_map, auto) qed

lemma mctxt_fill_partially_fills_length [simp]:
  assumes "sublist ss ts"
  shows "length (mctxt_fill_partially_fills ss ts) = length ts"
  using assms proof (induct rule: sublist_induct2, insert length_map, auto) qed

lemma mctxt_fill_partially_numholes:
  assumes "sublist ss ts"
  shows "listsum [num_holes ci . ci \<leftarrow> mctxt_fill_partially_mctxts ss ts] = length ss"
proof (induct ss ts rule: sublist_induct2)
  case (goal3 s ss ts) 
  have ls_one: "\<And> as. listsum (1 # as) = listsum as + Suc 0" 
    by (metis One_nat_def Suc_eq_plus1 Suc_eq_plus1_left listsum_simps(2))
  from goal3 show ?case 
    unfolding mctxt_fill_partially_mctxts.simps list.size
    by (metis (full_types) One_nat_def Suc_eq_plus1 ls_one list.map(2) num_holes.simps(2))
next
  case (goal4 s ss t ts) 
  have ls_zero: "\<And> as. listsum (0 # as) = listsum as" by (metis listsum_simps(2) monoid_add_class.add.left_neutral) 
  have else: 
    "mctxt_fill_partially_mctxts (s # ss) (t # ts) = term_to_mctxt t # mctxt_fill_partially_mctxts (s # ss) ts" 
    using goal4(1) by auto
  show ?case 
    unfolding else list.map(2) term_to_mctxt_numholes ls_zero goal4(5) ..
qed (auto iff: list.map term_to_mctxt_numholes assms)

lemma mctxt_fill_partially_sound:
  assumes sl: "sublist ss ts"
  shows "\<And> i. i < length ts \<Longrightarrow> ts!i =f (mctxt_fill_partially_mctxts ss ts ! i, mctxt_fill_partially_fills ss ts ! i)"
proof (rule eqfI)
  let ?zipped= "zip (mctxt_fill_partially_mctxts ss ts) (mctxt_fill_partially_fills ss ts)"

  have l: "length ?zipped = length ts" 
    unfolding length_zip mctxt_fill_partially_mctxts_length[OF sl] mctxt_fill_partially_fills_length[OF sl] by auto

  have fh: "ts = map (\<lambda> (ci,tsi) . fill_holes ci tsi) ?zipped" 
  proof (induct ss ts rule: sublist_induct2)
    case (goal2 ts) then show ?case
    proof (induct ts, insert term_to_mctxt_fill_holes, auto) qed
  qed (insert sl, auto)

  have nh: "list_all (\<lambda> (ci,tsi) . num_holes ci = length tsi) ?zipped"
  proof (induct ss ts rule: sublist_induct2)
    case (goal2 ts) then show ?case
    proof (induct ts, insert term_to_mctxt_numholes, auto) qed
  qed (insert sl, auto)

  { 
    fix i
    assume "i < length ts"
    then have 
      i1: "i < length (mctxt_fill_partially_mctxts ss ts)" and
      i2: "i < length (mctxt_fill_partially_fills ss ts)" 
      unfolding mctxt_fill_partially_mctxts_length[OF sl] mctxt_fill_partially_fills_length[OF sl] by auto
  } note i = this

  case goal1 
  then show ?case using fh nth_zip[OF i(1) i(2)]
    by (metis (lifting, no_types) goal1 list_update_id list_update_same_conv map_update split_conv)
  case goal2 then show ?case using nh[unfolded list_all_length] nth_zip[OF i(1) i(2)] 
    by (metis (lifting, full_types) l split_conv)
qed


lemma mctxt_fill_partially: 
  assumes  
  ss: "sublist ss ts" and
  t: "t =f (c,ts)"
  shows "\<exists> d. t =f (d,ss)"
proof - 
  let ?ds = "mctxt_fill_partially_mctxts ss ts"
  let ?sss = "mctxt_fill_partially_fills ss ts"

  have "fill_holes c ts =f (fill_holes_mctxt c ?ds, concat ?sss)"
    using 
      fill_holes_mctxt_sound eqfE(2)[OF t,symmetric] mctxt_fill_partially_sound[OF ss] 
      mctxt_fill_partially_mctxts_length[OF ss] mctxt_fill_partially_fills_length[OF ss]
    by metis
  also have "concat ?sss = ss" proof (induct ss ts rule: sublist_induct2, insert ss, auto) qed
  ultimately show ?thesis by (metis eqfE(1) t)
qed

end

